!=========================================================================
! Copyright (C) GemTalk Systems 1986-2016.  All Rights Reserved.
!
! $Id: identityindex.gs,v 1.8 2008-01-09 19:02:34 stever Exp $
!
! Superclass Hierarchy:
!   IdentityIndex, Array, SequenceableCollection, Collection, Object.
!
! class created in idxclasses.topaz
!
!=========================================================================

removeallmethods IdentityIndex
removeallclassmethods IdentityIndex
protectmethods

category: 'For Documentation Installation only'
classmethod: IdentityIndex
installDocumentation

< unprotected >

| doc |
doc := GsClassDocumentation _newForPrivateGsClass: self.

doc documentClassWith: (GsDocText new).

self description: doc.
%

! ------------------- Instance methods for IdentityIndex
category: 'Error Handling'
method: IdentityIndex
_errorCannotInvokeRangeOperationOnIdentityIndex

"An attempt was made to use a range operation (< > = <= >=) on a path
 expression only supported by an identity index."

<protected>

^ self _error: #rtErrIdentityIndexCannotInvokeRangeOperation
%

category: 'Searching'
method: IdentityIndex
_findFirstUnsharedPathTerm

"Returns the first path term that is not shared by any other indexes.  This is
 the first path term whose only index is the receiver.  If one is not found
 (the index shares the same path with another index), then returns nil."

<protected>

| pathTerm sz indexObj |

sz := self size.
1 to: sz do: [ :i |
    pathTerm := self at: i.
    ( pathTerm size == 1 _and: [ self == (pathTerm at: 1) ] )
        ifTrue: [ ^ pathTerm ]
].
" check if there is an equality and identity index on the same path "
pathTerm := self at: self size.
" for each index that utilizes the last path term "
1 to: pathTerm size do: [ :i |
    indexObj := pathTerm at: i.
    " if some other index has the same number of path terms "
    ( indexObj ~~ self _and: [ indexObj size == sz ] )
        ifTrue: [ ^ pathTerm ]
].
^ nil
%

category: 'Searching'
method: IdentityIndex
_findSetValuedPathTermOffset

"Returns the offset of the set-valued path term in the receiver.  If none
 exists, returns 0."

<protected>

1 to: self size do: [ :i |
    (self at: i) indicatesNsc ifTrue: [ ^ i ]
].
^ 0
%

category: 'Updating Indexes'
method: IdentityIndex
changingSizeOfByteObject: aByteObject to: newSize

"The size of the byte object is being changed.  For identity indexes, nothing
 is required to keep the index up to date."

<protected>
%

category: 'Query Select'
method: IdentityIndex
findAllValuesEqualTo: aValue

"Identity indexes do not support equality comparison."

<protected>

^ self _errorCannotInvokeRangeOperationOnIdentityIndex
%

category: 'Query Select'
method: IdentityIndex
findAllValuesGreaterThanKey: aValue andEquals: aBoolean

"Identity indexes do not support range comparison."

<protected>

^ self _errorCannotInvokeRangeOperationOnIdentityIndex
%

category: 'Query Select'
method: IdentityIndex
findAllValuesIdenticalTo: aValue

"Returns a set of all values that satisfy the query."

<protected>

| tmpHolder startPt |

" optimize query for objects in the NSC itself "
self isIndexOnRootNsc
    ifTrue: [ ^ self _allIdenticalTo: aValue in: nscRoot ].

tmpHolder := NscBuilder for: nscRoot speciesForSelect new max: self sizeForNscBuilder.

startPt := self lastPathComponentsDictionaryOffset.
self _addAllFor: aValue
    into: tmpHolder
    offset: startPt
    pathTerm: (self at: startPt).

^ tmpHolder completeBag
%

category: 'Query Select'
method: IdentityIndex
findAllValuesLessThanKey: aValue andEquals: aBoolean

"Identity indexes do not support range comparison."

<protected>

^ self _errorCannotInvokeRangeOperationOnIdentityIndex
%

category: 'Query Select'
method: IdentityIndex
findAllValuesNotEqualTo: aValue

"Identity indexes do not support equality comparison."

<protected>

^ self _errorCannotInvokeRangeOperationOnIdentityIndex
%

category: 'Query Select'
method: IdentityIndex
findAllValuesNotIdenticalTo: aValue

"Returns a set of all values that satisfy the query."

<protected>

| tmpHolder |

self isIndexOnRootNsc
    ifTrue: [ 
        tmpHolder := nscRoot copy.
        tmpHolder removeAll: (self _allIdenticalTo: aValue in: nscRoot).
        ^ tmpHolder
    ].

self hasSetValuedTerm
  ifTrue: [ ^ self _findAllValuesThroughSetsNotIdenticalTo: aValue ].

tmpHolder := NscBuilder for: nscRoot speciesForSelect new max: self sizeForNscBuilder.

" get all objects identical to aValue and identical to nil "
tmpHolder _addAllNsc: (self findAllValuesIdenticalTo: aValue) _asIdentityBag.
tmpHolder _addAllNsc: (self _allIdenticalTo: nil in: nscRoot).

" get all objects with nil along the path "
1 to: self size - 1 do: [ :i |
    self _addAllFor: nil
        into: tmpHolder
        offset: i
        pathTerm: (self at: i)
].

^ nscRoot _asIdentityBag - tmpHolder completeBag _asIdentityBag.
%

category: 'Query Detect'
method: IdentityIndex
findFirstValueEqualTo: aValue

"Identity indexes do not support range comparison."

<protected>

^ self _errorCannotInvokeRangeOperationOnIdentityIndex
%

category: 'Query Detect'
method: IdentityIndex
findFirstValueGreaterThanKey: aValue
andEquals: aBoolean

"Identity indexes do not support range comparison."

<protected>

^ self _errorCannotInvokeRangeOperationOnIdentityIndex
%

category: 'Query Detect'
method: IdentityIndex
findFirstValueIdenticalTo: aValue

"Returns the first value that satisfies the query."

<protected>

" optimize query for objects in the NSC itself "
self isIndexOnRootNsc
    ifTrue: [
        (nscRoot includesIdentical: aValue)
            ifTrue: [ ^ aValue ]
            ifFalse: [ ^ #_incompletePathTraversal ]
    ].

^ self _findFirstValueForKey: aValue
%

category: 'Query Detect'
method: IdentityIndex
findFirstValueLessThanKey: aValue
andEquals: aBoolean

"Identity indexes do not support range comparison."

<protected>

^ self _errorCannotInvokeRangeOperationOnIdentityIndex
%

category: 'Query Detect'
method: IdentityIndex
findFirstValueNotEqualTo: aValue

"Identity indexes do not support range comparison."

<protected>

^ self _errorCannotInvokeRangeOperationOnIdentityIndex
%

category: 'Accessing'
method: IdentityIndex
firstPathTerm

"Returns the first path term in the path components list."

<protected>

^ self at: 1
%

category: 'Testing'
method: IdentityIndex
hasIndexOnPath: pathArray

"Returns whether the receiver's path components match the path represented by
 pathArray."

<protected>

self size == pathArray size
    ifFalse: [ ^ false ].

1 to: self size do: [ :i |
    (self at: i) name == (pathArray at: i)
        ifFalse: [ ^ false ]
].
^ isComplete
%

category: 'Testing'
method: IdentityIndex
hasIndexOnPath: pathArray startingAt: offset

"Returns whether the receiver's path components match the path represented by
 pathArray."

<protected>

self size - offset + 1 == pathArray size
  ifFalse: [ ^ false ].

offset to: self size do: [ :i |
  (self at: i) name == (pathArray at: i + 1 - offset)
    ifFalse: [ ^ false ].
].
^ isComplete
%

category: 'Testing'
method: IdentityIndex
hasSetValuedTerm

"Returns true if the receiver has any path term that indicates a set-valued
 instance variable."

<protected>

1 to: self size do: [ :i |
    (self at: i) indicatesNsc ifTrue: [ ^ true ]
].
^ false
%

category: 'Accessing'
method: IdentityIndex
indexDictionary

"Returns the value of the instance variable 'indexDictionary'."

<protected>

^indexDictionary
%

category: 'Updating'
method: IdentityIndex
indexDictionary: newValue

"Modifies the value of the instance variable 'indexDictionary'."

<protected>

indexDictionary := newValue
%

category: 'Testing'
method: IdentityIndex
isIdentityIndex

"Returns true."

<unprotected>

^ true
%

category: 'Testing'
method: IdentityIndex
isIndexOnNscElements

"Returns whether the receiver is an index directly on the elements of an NSC
 (either the root NSC or a set-valued instance variable on the path)."

<protected>

^ self lastPathTerm indicatesIndexOnNscElements
%

category: 'Testing'
method: IdentityIndex
isIndexOnRootNsc

"Returns whether the receiver is an index directly on the elements of
 the root NSC."

<protected>

^ self lastPathTerm indicatesIndexOnRootNsc
%

category: 'Testing'
method: IdentityIndex
isPathEvaluator

"Returns false."

<unprotected>

^ false
%

category: 'Testing'
method: IdentityIndex
isRangeEqualityIndex

"Returns false."

<unprotected>

^ false
%

category: 'Accessing'
method: IdentityIndex
lastPathComponentsDictionaryOffset

"Returns the last index into the path components list that is mapped in the
 index dictionary (as opposed to the B-tree)."

<protected>

^ self size
%

category: 'Accessing'
method: IdentityIndex
lastPathTerm

"Returns the last path term in the path components list."

<protected>

^ self at: self size
%

category: 'Accessing'
method: IdentityIndex
nscRoot

"Returns the value of the instance variable 'nscRoot'."

<protected>

^nscRoot
%

category: 'Updating'
method: IdentityIndex
nscRoot: newValue

"Modify the value of the instance variable 'nscRoot'."

<protected>

nscRoot := newValue
%

category: 'Accessing'
method: IdentityIndex
pathComponentsString

"Returns a string of the path components."

<protected>

^ self pathComponentsStringStartingAt: 1
%

category: 'Repository Conversion'
method: IdentityIndex
pathComponentsStringVersion: versionNum

"This method is a kludge.  The right way would be to have corresponding
 obsolete classes because instance variables were added going from 4.1 to 5.0."

<protected>

(versionNum = 50)
  ifTrue: [ ^ self pathComponentsString ].

(versionNum = 41)
  ifTrue: [ ^ self pathComponentsString41 ].

self halt: 'Invalid version num specified.'
%

category: 'Repository Conversion'
method: IdentityIndex
pathComponentsString41

"This method is a kludge.  The right way would be to have corresponding
 obsolete classes because instance variables were added going from 4.1 to 5.0."

<protected>

| objSize newIdIndex numInstVars |

newIdIndex := IdentityIndex new.

numInstVars := IdentityIndex allInstVarNames size.
objSize := self _primitiveSize.

(objSize > numInstVars)
  ifTrue: [
    4 to: numInstVars do: [ :i |
      newIdIndex add: (self instVarAt: i).
      ].

    1 to: self size do: [ :i |
      newIdIndex add: (self at: i).
      ].
    ]
  ifFalse: [
    4 to: objSize do:  [ :i |
      newIdIndex add: (self instVarAt: i).
      ].
    ].

^ newIdIndex pathComponentsString.
%

category: 'Accessing'
method: IdentityIndex
pathComponentsStringStartingAt: offset

"Returns a string of the path components."

<protected>

| str sz |

sz := self size.
str := String new.
offset to: sz do: [ :i |
    str addAll: (self at: i) name.
    i == sz
        ifFalse: [ str add: $.  ]
].
^ str
%

category: 'Accessing'
method: IdentityIndex
pathSize

"Returns the number of path terms."

<protected>

^ self size
%

category: 'Traversing'
method: IdentityIndex
traverse: anObject upTo: aPathTerm

"Traverse the sub-objects of the given object using the path terms of the
 receiver up to the given path term.  This method assumes that the index is not
 over a set-valued instance variable.  For that kind of index, use the
 traverse:upTo:startingAt:into: method."

<protected>

| nextObj ivOffset sz i|

nil == anObject
    ifTrue: [ ^ anObject ].
nextObj := anObject.
sz := self size.
i := 1.
[ (self at: i) ~~ aPathTerm _and: [ i <= sz ] ] whileTrue: [
    ivOffset := nextObj _findIVOffsetForPathTerm: (self at: i).
    ivOffset == nil
        ifTrue: [
            nextObj _errorPathTermNotInDependencyList: (self at: i)  .
            nextObj := nil
            ]
        ifFalse:[ nextObj := nextObj instVarAt: ivOffset ] .
    nil == nextObj
        ifTrue: [ ^ nextObj ].
    i := i + 1
].
^ nextObj
%

category: 'Traversing'
method: IdentityIndex
traverse: anObject
upTo: aPathTerm
startingAt: anOffset
into: resultSet
addingIndex: addingIndex

"Traverse the sub-objects of anObject, using the path terms of the receiver up
 to the given path term.  Add each object at the end of the traversal to the
 result set.  If addingIndex is true, add an index list entry for the
 receiver to any NSCs encountered along the path; if false, remove an index
 list entry for the receiver.  Returns the result set."

<protected>

| nextObj ivOffset nscOffset sz i |

nextObj := anObject.
sz := self size.
i := anOffset.
[ (self at: i) ~~ aPathTerm _and:
[ (i <= sz) _and:
[ nil ~~ nextObj ] ] ] whileTrue: [

    " if indexing over a set-valued instance variable "
    (self at: i) indicatesNsc
        ifTrue: [
            nscOffset := i + 1.
            nextObj _indexedPaths ~~ nil
                ifTrue: [
                    addingIndex
                        ifTrue: [
                            nextObj _indexedPaths addIndex: self withOffset: (self at: i) offset + 1 nsc: nextObj
                        ]
                        ifFalse: [
                            nextObj _indexedPaths removeIndex: self withOffset: nscOffset for: nextObj
                        ]
                ].
            " for each object in the set-valued instance variable "
            nextObj do: [ :obj |
                " make recursive call "
                self
                    traverse: obj
                    upTo: aPathTerm
                    startingAt: nscOffset
                    into: resultSet
                    addingIndex: addingIndex
            ].
            ^ resultSet
        ]
        ifFalse: [
            " get the next object along the path "
            ivOffset := nextObj _findIVOffsetForPathTerm: (self at: i).
            ivOffset == nil
              ifTrue: [
                nextObj _errorPathTermNotInDependencyList: (self at: i) .
                nextObj := nil
                ]
              ifFalse:[ nextObj := nextObj instVarAt: ivOffset ] .
        ].
    i := i + 1
].
( nil ~~ nextObj _and: [ resultSet ~~ nil ] )
    ifTrue: [ resultSet add: nextObj ].

^ resultSet
%

category: 'Traversing'
method: IdentityIndex
traverseObject: anObject

"Traverse the sub-objects of the given object using the path terms of the
 receiver.  This method assumes that the index is not over a set-valued
 instance variable.  For that kind of index, use the traverse:startingAt:into:
 method.  If a nil value is reached before the end of the path, returns the
 incompletePathTraversal object."

<protected>

self isIndexOnNscElements
    ifTrue: [ ^ anObject ].

nil == anObject
    ifTrue: [ ^ #_incompletePathTraversal ].

^ self _traverseObject: anObject
%

category: 'Updating Indexes'
method: IdentityIndex
updatingByteObject: aByteObject startingAt: startPt withNewValue: newValue

"The byte object is being updated.  For identity indexes, nothing is required
 to keep the index up to date."

<unprotected>
%

category: 'Testing'
method: IdentityIndex
checkAuthorization

"Determine whether the current user has read authorization on all segments for
 each path term utilized by the receiver.  If not, raise an authorization
 error."

<protected>

| segs |

1 to: self size do: [ :i |
    segs := (self at: i) segments.
    System _addEntireObjectToRcReadSet: segs.
    1 to: segs size do: [ :i |
        " this will determine if we can read the segment "
        (segs _at: i) currentUserCanRead
            ifFalse: [
                self
                    _error: #rtErrIndexAuthErrSegRead
                    args: #[ self pathComponentsString, segs _at: i ].
                self _uncontinuableError
            ]
    ]
].
%

category: 'Traversing'
method: IdentityIndex
_traverseObject: anObject

"Traverse the sub-objects of the given object using the path terms of the
 receiver.  This method assumes that the index is not over a set-valued
 instance variable.  For that kind of index, use the traverse:startingAt:into:
 method.  If a nil value is reached before the end of the path, returns the
 incompletePathTraversal object."

<protected>

| nextObj ivOffset sz pathTerm |

nextObj := anObject.
sz := self size.
1 to: sz do: [ :i |
    pathTerm := self at: i.
    ivOffset := nextObj _findIVOffsetForPathTerm: pathTerm.
    ivOffset == nil
        ifTrue: [
           nextObj _errorPathTermNotInDependencyList: pathTerm .
           nextObj := nil
           ]
        ifFalse:[ nextObj := nextObj instVarAt: ivOffset ] .

    ( nil == nextObj _and: [ i ~= sz ] )
        ifTrue: [ ^ #_incompletePathTraversal ].
].
^ nextObj
%

category: 'Updating Indexes'
method: IdentityIndex
addDirectMappingsFor: aBag
indexList: iList
updateIndexListSegments: aBoolean
commitInterval: anInteger

"Add an entry for an index on the elements of the NSC."

<protected>

| anObject firstPathTerm |
firstPathTerm := self firstPathTerm.
anInteger ~~ nil
  ifTrue: [ nscRoot _lockForIndexCreation ].

aBoolean
  ifTrue: [
    1 to: aBag size do: [ :i |
      anInteger ~~ nil
        ifTrue: [
          nscRoot _commitIndexCreation: self at: i commitInterval: anInteger
        ].

      anObject := aBag _at: i.
      " insert an entry into the index dictionary "
      indexDictionary
        _at: anObject
        put: anObject
        term: firstPathTerm
        logging: false.
      iList addSegment: anObject segment logging: false
    ]
  ]
  ifFalse: [
    1 to: aBag size do: [ :i |
      anInteger ~~ nil
        ifTrue: [
          nscRoot _commitIndexCreation: self at: i commitInterval: anInteger
        ].

      anObject := aBag _at: i.
      " insert an entry into the index dictionary "
      indexDictionary
        _at: anObject
        put: anObject
        term: firstPathTerm
        logging: false
    ]
  ]
%

category: 'Updating Indexes'
method: IdentityIndex
preIndexCreation

"Initializes the working set of collision buckets in the index dictionary."

<unprotected>

isComplete := false.
progress := 0.

indexDictionary ~~ nil
  ifTrue: [ indexDictionary initializeWorkingSetInterval: self ].
^ nil
%

category: 'Updating Indexes'
method: IdentityIndex
postIndexCreation: original commitInterval: anInteger

"Iterates through all ID entry holders, updating the index dictionary."

<unprotected>

" indicate that it is ok to update the dictionary "
indexDictionary ~~ nil
  ifTrue: [ | array |
    " get the Array of IndexDictionaryEntryHolder's "
    array := indexDictionary entryHolders.
    1 to: array size do: [ :i | | holder |
      holder := array at: i.
      holder update.
      holder size: 0.
    ].
    " nil out the Array of ID entry holders for garbage collection "
    indexDictionary entryHolders: nil.
    indexDictionary okToUpdate: true.
  ].

isComplete := true.

anInteger ~~ nil
  ifTrue: [
    System commitTransaction
      ifFalse: [ nscRoot _errorCouldNotCommitDuringIndexCreation ].
    System transactionMode == #manualBegin ifTrue:
        [System beginTransaction]
  ]
%

category: 'Searching'
method: IdentityIndex
_findFirstValueForKey: aKey

"Returns any root value for the given key by following the path terms in reverse
 order.  If none are found, returns aValue."

<protected>

| startPt tmpVal incomplete |

nil == indexDictionary
  ifTrue: [ ^ aKey ].

tmpVal := aKey.
self isIndexOnNscElements
    ifTrue: [ startPt := self size ]
    ifFalse: [ startPt := self lastPathComponentsDictionaryOffset ].

incomplete := #_incompletePathTraversal.
startPt downTo: 1 do: [ :i |
  tmpVal := indexDictionary
    firstAt: tmpVal
    term: (self at: i)
    otherwise: incomplete.
  (tmpVal == incomplete)
    ifTrue: [ ^ incomplete ].
].
^ tmpVal
%

category: 'Accessing'
method: IdentityIndex
at: anIndex

"Protected version of Object | at:."

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [^ self _errorNotIndexable].
(anIndex _isSmallInteger)
  ifTrue: [^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [^ self _errorNonIntegerIndex: anIndex].
self _primitiveFailed: #at: .
self _uncontinuableError
%

category: 'Accessing'
method: IdentityIndex
instVarAt: anIndex

"Protected version of Object | instVarAt:."

<protected primitive: 811"with rcvrClass">

| numInstVars |
(anIndex _isInteger)
ifTrue:
[
   numInstVars:= self class allInstVarNames size.
   (numInstVars == 0)
   ifTrue:
      [ ^ self _error: #rtErrNoInstVars].
   ((anIndex >= 1) & (anIndex <= numInstVars))
   ifFalse: "out of bounds"
      [ ^ self _errorIndexOutOfRange: anIndex]
]
ifFalse: "anIndex not an integer"
   [ ^ self _errorNonIntegerIndex: anIndex
].

self _primitiveFailed: #instVarAt: .
self _uncontinuableError
%

category: 'Accessing'
method: IdentityIndex
_at: anIndex

"Protected version of Object | _at:."

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
(anIndex _isSmallInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_at: .
self _uncontinuableError
%

category: 'Accessing'
method: IdentityIndex
_basicAt: anIndex

"Protected version of Object | _basicAt:."

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
(anIndex _isSmallInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_basicAt: .
self _uncontinuableError
%

category: 'Accessing'
method: IdentityIndex
_primitiveAt: anIndex

"Protected version of Object | _primitiveAt:."

<protected primitive: 89>

(anIndex _isInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_primitiveAt: .
self _uncontinuableError
%

category: 'Updating'
method: IdentityIndex
at: anIndex put: aValue

"Protected version of Object | at:put:."

<protected primitive: 268>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue: [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #at:put: .
self _uncontinuableError
%

category: 'Updating'
method: IdentityIndex
instVarAt: anIndex put: aValue

"Protected version of Object | instVarAt:put:."

<protected primitive: 804"with rcvrClass">

(self class allInstVarNames size == 0)
  ifTrue: [ ^ self _error: #rtErrNoInstVars].
(anIndex _isInteger)
  ifTrue: [((anIndex < 1) | (anIndex > self class allInstVarNames size))
        ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]] "out of bounds"
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #instVarAt:put: .
self _uncontinuableError
%

category: 'Updating'
method: IdentityIndex
size: anInteger

"Protected version of Object | size:."

<protected primitive: 803"with rcvrClass">

(self class isIndexable not) "not an indexable object"
  ifTrue: [ ^ self _errorNotIndexable].
(anInteger _isInteger)
   ifTrue: [ ^ self _errorIndexOutOfRange: anInteger]
   ifFalse: [ ^ self _errorNonIntegerIndex: anInteger].
self _primitiveFailed: #size: .
self _uncontinuableError
%

category: 'Updating'
method: IdentityIndex
_basicAt: anIndex put: aValue

"Protected version of Object | _basicAt:put:."

<protected primitive: 268>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_basicAt:put: .
self _uncontinuableError
%

category: 'Updating'
method: IdentityIndex
_at: anIndex put: aValue

"Protected version of Object | _at:put:."

<protected primitive: 268>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_at:put: .
self _uncontinuableError
%

category: 'Updating'
method: IdentityIndex
_basicSize: anInteger

"Protected version of Object | _basicSize:."

<protected primitive: 803"with rcvrClass">

(self class isIndexable not) "not an indexable object"
  ifTrue: [ ^ self _errorNotIndexable].
(anInteger _isInteger)
   ifTrue: [ ^ self _errorIndexOutOfRange: anInteger]
   ifFalse: [ ^ self _errorNonIntegerIndex: anInteger].
self _primitiveFailed: #size: .
self _uncontinuableError
%

category: 'Updating'
method: IdentityIndex
_primitiveAt: anIndex put: aValue

"Protected version of Object | _primitiveAt:put:."

<protected primitive: 281>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
((anIndex < 1) | (anIndex > (self _primitiveSize + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_primitiveAt:put: .
self _uncontinuableError
%

category: 'Reduced Conflict Support'
method: IdentityIndex
_rcAt: anIndex

"Protected version of Object | _rcAt:."

<protected primitive: 92>

self _primitiveFailed: #_rcAt: .
self _uncontinuableError
%

category: 'Accessing'
method: IdentityIndex
basicAt: anIndex

"Protected version of Object | basicAt:."

<protected>

^self _basicAt: anIndex
%

category: 'Updating'
method: IdentityIndex
basicAt: anIndex put: aValue

"Protected version of Object | basicAt:put:."

<protected>

^self _basicAt: anIndex put: aValue
%

category: 'Copying'
method: IdentityIndex
copyFrom: index1 to: index2 into: aSeqColl startingAt: destIndex

"Protected version of Array | copyFrom:to:into:startingAt:."

<protected primitive: 808"with rcvrClass">

^ super copyFrom: index1 to: index2 into: aSeqColl startingAt: destIndex
%

category: 'Copying'
method: IdentityIndex
removeFrom: startIndex to: stopIndex

"Protected version of Array | removeFrom:to:."

<protected primitive: 801>

^ super removeFrom: startIndex to: stopIndex
%

category: 'Removing'
method: IdentityIndex
_deleteNoShrinkFrom: startIndex to: stopIndex

"Protected version of Array | _deleteNoShrinkFrom:to:."

<protected primitive: 802 "with rcvrClass">

^ super _deleteNoShrinkFrom: startIndex to: stopIndex
%

category: 'Copying'
method: IdentityIndex
_insertAt: destOffset
from: anArray
fromStart: startOffset
fromEnd: endOffset
numToMoveDown: numToMove

"Protected version of Array | #_insertAt:from:fromStart:fromEnd:numToMoveDown:."

<protected primitive: 805"with rcvrClass">

^ super _insertAt: destOffset from: anArray fromStart: startOffset
            fromEnd: endOffset numToMoveDown: numToMove
%

category: 'Copying'
method: IdentityIndex
insertAll: aSeqColl at: anIndex

"Protected version of Array | insertAll:at:."

<protected primitive: 806"with rcvrClass">

^ super insertAll: aSeqColl at: anIndex
%

category: 'Adding'
method: IdentityIndex
add: newObject

"Protected version of Array | add:."

<protected primitive: 203>

self _primitiveFailed: #add: .
self _uncontinuableError
%

category: 'Adding'
method: IdentityIndex
addLast: newObject

"Protected version of Array | addLast:."

<protected primitive: 203>

self _primitiveFailed: #addLast: .
self _uncontinuableError
%

category: 'Adding'
method: IdentityIndex
_addAllFromNsc: aBag

"Protected version of Array | _addAllFromNsc:."

<protected primitive: 202>

self _primitiveFailed: #_addAllFromNsc: .
self _uncontinuableError
%

category: 'Query Detect'
method: IdentityIndex
_findFirstKeyNotIdenticalTo: aKey atTerm: pathTerm

"Find the first entry with the given path term whose key is not identical to
 the given key and returns the entry's key."

<protected>

^ indexDictionary _findFirstKeyNotIdenticalTo: aKey atTerm: pathTerm
%

category: 'Query Detect'
method: IdentityIndex
findFirstValueNotIdenticalTo: aValue

"Returns the first value that satisfies the query."

<protected>

| key lastPathTerm |

lastPathTerm := self lastPathTerm.
" optimize query for objects in the NSC itself "
self isIndexOnRootNsc
  ifTrue: [
    1 to: nscRoot size do: [ :i |
      ((nscRoot _at: i) == aValue)
        ifFalse: [ ^ nscRoot _at: i ]
    ].
    ^ #_incompletePathTraversal
  ].

key := self _findFirstKeyNotIdenticalTo: aValue atTerm: lastPathTerm.
(key == #_incompletePathTraversal)
  ifTrue: [ ^ #_incompletePathTraversal ].

^ self findFirstValueIdenticalTo: key
%

category: 'Query Support'
method: IdentityIndex
asPathEvaluator

"Returns a PathEvaluator over the same path as the receiver."

<protected>

| pathEval |

pathEval := PathEvaluator new.

1 to: self size do: [ :i |
    pathEval add: (self at: i) name
].
pathEval nsc: nscRoot.
^ pathEval asMostSpecificType
%

category: 'Query Support'
method: IdentityIndex
_addAllFor: aKey
into: anNscBuilderOrArray
offset: offset
pathTerm: pathTerm

"Traverse the reverse mappings for the given key, adding the object at the end
 of the traversal to the given NscBuilder or Array."

<protected primitive: 328>

self _primitiveFailed: #_addAllFor:into:offset:pathTerm: .
self _uncontinuableError
%

category: 'Updating'
method: IdentityIndex
_setPathTermState

"For each path term, indicate the need to update the index dictionary."

<protected>

1 to: self size do: [ :i |
    (self at: i) updateDict: indexDictionary.
]
%

category: 'Query Support'
method: IdentityIndex
_allIdenticalTo: anObject in: nsc

"Returns a new NSC containing the given object with the number of occurrences
 as in the root NSC."

<protected>

| result tmp |
result := self _primAllIdenticalTo: anObject in: nsc _asIdentityBag.
nsc _isIdentityBag
  ifFalse: [
    tmp := nsc class new: result size.
    tmp addAll: result.
    result := tmp
  ].
^ result
%

category: 'Query Support'
method: IdentityIndex
_primAllIdenticalTo: anObject in: nsc

"Returns a new NSC containing the given object with the number of occurrences
 as in the root NSC."

<protected primitive: 340>

self _primitiveFailed: #allIdenticalTo:in: .
self _uncontinuableError
%

category: 'Query Support'
method: IdentityIndex
hasSamePathAs: anIndex

"Returns whether the given index is on the same path as the receiver.
 This can only be true if the index is identical."

<protected>

^ self == anIndex
%

category: 'Indexing Support'
method: IdentityIndex
_partialPathComponentsStringUpTo: offset

"Returns the path components string up to the given offset."

<protected>

| str |
str := String new.
1 to: (offset min: self size) do: [ :i |
    str add: (self at: i) name.
    i == offset
        ifFalse: [ str add: $. ]
].
^ str
%

category: 'Traversing'
method: IdentityIndex
_traverseObject: anObject
incompletesInto: incompleteArray
incomplete: incomplete

"Traverse the sub-objects of the given object using the path terms of the
 receiver.  This method assumes that the index is not over a set-valued
 instance variable.  For that kind of index, use the traverse:startingAt:into:
 method.  If a nil value is reached before the end of the path, place the object
 into the appropriate Array in the incompleteArray of Arrays."

<protected>

| nextObj ivOffset sz pathTerm |

nextObj := anObject.
sz := self size.
1 to: sz do: [ :i |
    pathTerm := self at: i.
    ivOffset := nextObj _findIVOffsetForPathTerm: pathTerm.
    ivOffset == nil
        ifTrue: [
           nextObj _errorPathTermNotInDependencyList: pathTerm .
           nextObj := nil
           ]
        ifFalse:[ nextObj := nextObj instVarAt: ivOffset ] .

    ( nil == nextObj _and: [ i ~= sz ] )
        ifTrue: [
            incompleteArray == nil
                ifTrue: [ ^ nil ].

            (incompleteArray at: i +1) add: anObject.
            ^ incomplete
        ].
].
^ nextObj
%

category: 'Traversing'
method: IdentityIndex
traverseObject: anObject
incompletesInto: incompleteArray
incomplete: incomplete

"Traverse the sub-objects of the given object using the path terms of the
 receiver.  This method assumes that the index is not over a set-valued
 instance variable.  For that kind of index, use the traverse:startingAt:into:
 method.  If a nil value is reached before the end of the path, place the object
 into the appropriate Array in the incompleteArray of Arrays."

<protected>

self isIndexOnNscElements
    ifTrue: [ ^ anObject ].

nil == anObject
    ifTrue: [ ^ incomplete ].

^ self _traverseObject: anObject
  incompletesInto: incompleteArray
  incomplete: incomplete
%

category: 'Accessing'
method: IdentityIndex
lastElementClass

" "

<protected>

^ nscRoot getLastElementConstraintOnPath: self pathComponentsString
%

category: 'Query Support'
method: IdentityIndex
_findAllValuesThroughSetsNotIdenticalTo: aKey

"Returns a set of all values that satisfy the query.  This method works
 when the path is through a set-valued instance variable."

<protected>

| tmpVals tmpHolder startPt pathTerm |
tmpHolder := NscBuilder for: nscRoot speciesForSelect new max: self sizeForNscBuilder.

tmpVals := indexDictionary _findAllValuesNotIdenticalTo: aKey
    atTerm: (self at: self lastPathComponentsDictionaryOffset)
    into: Array new.

startPt := self lastPathComponentsDictionaryOffset - 1.
pathTerm := self at: startPt.

1 to: tmpVals size do: [ :i |
    self _addAllFor: (tmpVals _at: i)
        into: tmpHolder
        offset: startPt
        pathTerm: pathTerm
].
^ tmpHolder completeBag
%

category: 'Accessing'
method: IdentityIndex
sizeForNscBuilder

"Returns the size to be used when an NscBuilder is collecting objects
 based upon the receiver's path."

<protected>

^ self hasSetValuedTerm
  ifTrue: [ NscBuilder maxSize ]
  ifFalse: [ nscRoot size ]
%

category: 'Formatting'
method: IdentityIndex
printOn: aStream

"Puts a displayable representation of the receiver on the given stream."

"Copy the implementation from Object so we don't inherit it from Collection."

< unprotected >

aStream nextPutAll: self asString
%

category: 'Clustering'
method: IdentityIndex
clusterDepthFirst

"Cluster the receiver.  Only need to cluster the path terms."

<protected>

1 to: self size do: [ :i |
  (self at: i) clusterDepthFirst
]
%

category: 'Statistics'
method: IdentityIndex
_statisticsInto: dict

"Puts statistical information into the given dictionary."

<protected>

%

! Deleted:  _obj: anObject isIdenticalTo: anotherObject

category: 'Testing'
method: IdentityIndex
_isIndexObject

"Returns true if the receiver is an index object; returns false otherwise."

<unprotected>

^ true
%

category: 'Accessing'
method: IdentityIndex
progress

"Returns the value of the instance variable 'progress'."

<protected>

^progress
%

category: 'Testing'
method: IdentityIndex
isComplete

"Returns the value of the instance variable 'isComplete'."

<protected>

isComplete == nil
  ifTrue: [ ^ true ].
^ isComplete
%

category: 'Accessing'
method: IdentityIndex
nwayMergeProgress

"Returns nil (identity indexes do not utilize n-way merge)."

<protected>

^ nil
%

category: 'Updating'
method: IdentityIndex
progress: newValue

"Modifies the value of the instance variable 'progress'."

<protected>

progress := newValue
%

category: 'Updating'
method: IdentityIndex
isComplete: newValue

"Modifies the value of the instance variable 'isComplete'."

<protected>

isComplete := newValue
%

unprotectmethods

